""" Utility functions for extracting COS spectral data into lightcurves

"""

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

from astropy.io import fits
import os
import numpy as np
from scipy.interpolate import interp1d
import six

from .utils import expand_refname

__all__ = ['extract',
           'collect_inputs',
           'get_both_filenames',
           'get_extraction_region'
           ]

#-------------------------------------------------------------------------------

def extract(filename, **kwargs):
    """ Extract lightcurve from COS dataset

    This is the main driver of the lightcuve extracion, and definitely
    needs some better documentation.

    Parameters
    ----------
    filename : str
        name of FITS file to extract from
    **kwargs : dict
        arbitrary keyword arguements for tailored extraction

    Returns
    -------
    data, meta : Astropy table, dict
        Table with extracted data and dictionary of metadata pairs
    """

    """
                if alt:
                    seg = input_hdus.keys()[0]
                    print('Filtering out times of SUN_ALT > {}'.format(alt))
                    print('--Total time in exposure: {}'.format(
                            input_hdus[seg][1].header['EXPTIME']))
                    try:
                        timeline = input_hdus[seg]['TIMELINE'].data
                        day_index = np.where(timeline['SUN_ALT'] > alt)

                        if len(day_index[0]):
                            bad_min = timeline['TIME'][day_index].min()
                            bad_max = timeline['TIME'][day_index].max()

                            print('--Removing times from {} to {} seconds'.format(bad_min, bad_max))
                            lc_index = np.where((times < bad_min) |
                                                (times > bad_max))

                            filter(lc_index)

                        else:
                            print('--No times matching criteria, no filtering performed')

                    except KeyError:
                        print('Warning: No TIMELINE extension found, make sure this dataset')
                        print('         was calibrated with the *_spt.fits file present')

    """

    input_files, input_hdus = collect_inputs(filename)
    input_headers = {}
    for segment in input_hdus:
        input_headers[segment] = {}
        for i, ext in enumerate(input_hdus[segment]):
            try:
                input_headers[segment][i] = ext.header._cards
            except AttributeError:
                pass

    verbosity = kwargs.get('verbosity', 0)
    step = kwargs.get('step', 1)
    wlim = kwargs.get('wlim', None)
    #-- If not specific wavlengths, truncate to good wavelengths
    #-- for each detector
    if not wlim:
        if fits.getval(filename, 'DETECTOR') == 'FUV':
            wlim = (915, 1800)
        else:
            wlim = (915, 3200)

    xlim = kwargs.get('xlim', (0, 16384))
    ylim = kwargs.get('ylim', None)
    filter_airglow = kwargs.get('filter_airglow', True)

    if fits.getval(filename, 'OBSTYPE') == 'IMAGING':
        print("Imaging observation found, resetting limits.")
        xlim = (0, 1024)
        ylim = (0, 512)
        wlim = (-1, 1)

    SECOND_PER_MJD = 1.15741e-5

    source_datasets = ','.join([os.path.basename(item) for item in input_files])
    meta = {'source': filename,
            'instrument' : 'COS',
            'headers': input_headers,
            'source_files': source_datasets,
            'stepsize': step,
            'wlim': wlim,
            'xlim': xlim,
            'ylim': ylim}

    if verbosity:
        print('#----------------------------------------#')
        print('Running LightCurve extraction for COS Data')
        print('#----------------------------------------#')
        print()
        print('Extracting from: {}'.format(input_files))
        print('With arguments:')
        print('With arguments:')
        print('step : {}'.format(step))
        print('wlim : {}'.format(wlim))
        print('xlim : {}'.format(xlim))
        print('ylim : {}'.format(ylim))
        print('filter_airglow : {}'.format(filter_airglow))
        print()

    #time = np.array([round(val, 3) for val in hdu[1].data['time']]).astype(np.float64)

    end = 0
    exptime = 0
    start = 0

    for segment, hdu in six.iteritems(input_hdus):
        start = max(start, hdu[1].data['TIME'].min())
        end = max(end, hdu[1].data['TIME'].max())
        exptime = max(exptime, hdu[1].header['EXPTIME'])

    if end > exptime:
        print("WARNING: data times go to {}, beyond exptime: {}".format(end, exptime))

    end = min(end, exptime)

    all_steps = np.arange(start, end+step, step)

    if all_steps[-1] > end:
        truncate = True
    else:
        truncate = False

    gross = 0
    flux = 0
    background = 0
    background_flux = 0

    for segment, hdu in six.iteritems(input_hdus):

        ### Source Calculation
        print('Segment {} fluxcal'.format(segment))
        if not ylim:
            ystart, yend = get_extraction_region(hdu, segment, 'spectrum')
        else:
            ystart, yend = ylim[0], ylim[1]

        meta['ylim{}'.format(segment.lower())] = (ystart, yend)

        if verbosity:
            print(xlim, ystart, yend, wlim, hdu[1].header['sdqflags'])
        index = extract_index(hdu,
                              xlim[0], xlim[1],
                              ystart, yend,
                              wlim[0], wlim[1],
                              hdu[1].header['sdqflags'],
                              filter_airglow=filter_airglow)
        if verbosity:
            print("{} #{} events".format(segment, len(index)))


        n_pixels = calc_npixels(hdu, index, xlim)

        gross += np.histogram(hdu['events'].data['time'][index],
                              all_steps,
                              weights=hdu['events'].data['epsilon'][index])[0]

        response_array = get_fluxes(hdu, index).mean()
        tds_corr = get_tds(hdu, index)

        weights = hdu['events'].data['epsilon'][index] / step / tds_corr / response_array
        flux +=  np.histogram(hdu['events'].data['time'][index], all_steps, weights=weights)[0] / n_pixels


        ### Background calculation
        #print('Background fluxcal')
        bstart, bend = get_extraction_region(hdu, segment, 'background1')
        index = extract_index(hdu,
                              xlim[0], xlim[1],
                              bstart, bend,
                              wlim[0], wlim[1],
                              hdu[1].header['sdqflags'],
                              filter_airglow=filter_airglow)

        bstart, bend = get_extraction_region(hdu, segment, 'background2')
        index = np.hstack( (index, extract_index(hdu,
                                                 xlim[0], xlim[1],
                                                 bstart, bend,
                                                 wlim[0], wlim[1],
                                                 hdu[1].header['sdqflags'],
                                                 filter_airglow=filter_airglow) ) )

        b_corr = ((bend - bstart) / (yend - ystart)) / 2.
        background += b_corr * np.histogram(hdu['events'].data['time'][index],
                                            all_steps,
                                            weights=hdu['events'].data['epsilon'][index] )[0]

        response_array = get_fluxes(hdu, index)
        tds_corr = get_tds(hdu, index)
        weights = hdu['events'].data['epsilon'][index] / step / tds_corr / response_array
        background_flux +=  (b_corr * np.histogram(hdu['events'].data['time'][index],
                                                   all_steps,
                                                   weights=weights)[0]) / n_pixels


    gross = gross
    flux = flux - background_flux
    background = background
    mjd = hdu[1].header['EXPSTART'] + np.array(all_steps[:-1]) * SECOND_PER_MJD
    bins = np.ones(len(gross)) * step
    times = all_steps[:-1]

    if truncate:
        if verbosity:
            print('Truncating the last event bin')

        gross = gross[:-1]
        flux = flux[:-1]
        background = background[:-1]
        mjd = mjd[:-1]
        bins = bins[:-1]
        times = times[:-1]

    data = {'dataset': np.ones(times.shape),
            'times': times,
            'mjd': mjd,
            'bins': bins,
            'gross': gross,
            'background': background,
            'flux': flux}

    if verbosity:
        print('Finished extraction for {}'.format(filename))
        print()

    return data, meta

#-------------------------------------------------------------------------------

def collect_inputs(filename):
    """Populate HDU dictionary from available corrtag files

    """

    with fits.open(filename) as hdu:
        if hdu[0].header['detector'] == 'NUV':
            all_filenames = [filename]

            out_hdu = fits.open(filename)
            if hdu[0].header['opt_elem'] == 'G230L':
                all_hdu = {'A': out_hdu, 'B': out_hdu}
            else:
                all_hdu = {'A': out_hdu, 'B': out_hdu, 'C': out_hdu}

        elif hdu[0].header['detector'] == 'FUV':
            file_a, file_b = get_both_filenames(filename)

            all_hdu = dict()
            all_filenames = []
            for name, letter in zip([file_a, file_b], ['A', 'B']):
                if os.path.exists(name):
                    all_filenames.append(name)
                    all_hdu[letter] = fits.open(name, ignore_missing_end=True)

    return all_filenames, all_hdu

#-------------------------------------------------------------------------------

def calc_npixels(hdu, index, xlim):
    try:
        max_xpix = round(min(hdu['events'].data['XCORR'][index].max(), xlim[1]))
        min_xpix = round(max(hdu['events'].data['XCORR'][index].min(), xlim[0]))
        n_pixels = (max_xpix - min_xpix) + 1
    except ValueError:
        n_pixels = 1

    return n_pixels

#-------------------------------------------------------------------------------

def extract_index(hdu, x_start, x_end,
                  y_start, y_end, w_start, w_end, sdqflags=0,
                  filter_airglow=True):
    """
    Extract event indeces from given HDU using input parameters.

    Wavelength regions containing geocoronal airglow emission will be excluded
    automatically.  These regions include Lyman Alpha between 1214 and 1217
    Angstroms and Oxygen I from 1300 to 1308.


    Parameters
    ----------
    hdu : HDUlist
        Header Data Unit from COS corrtag file
    x_start : float
        Lower bound of events in pixel space
    x_end : float
        Upper bound of events in pixel space
    y_start : float
        Lower bound of events in pixel space
    y_end : float
        Upper bound of events in pixel space
    w_start : float
        Lower bound of events in wavelength space
    w_end : float
        Upper bound of events in wavelength space
    sdqflags : int
        Bitwise DQ value of bad events

    Returns
    -------
    data_index : np.ndarray
        Indeces of desired events

    """

    if filter_airglow:
        lyman = (1208, 1225)
        oxygen = (1298, 1312)
    else:
        lyman = (w_end, w_start)
        oxygen = (w_end, w_start)

    data_index = np.where((hdu[1].data['XCORR'] >= x_start) &
                          (hdu[1].data['XCORR'] < x_end) &

                          (hdu[1].data['YCORR'] >= y_start) &
                          (hdu[1].data['YCORR'] < y_end) &

                          np.logical_not(hdu[1].data['DQ'] & sdqflags) &

                          ((hdu[1].data['WAVELENGTH'] > w_start) &
                           (hdu[1].data['WAVELENGTH'] < w_end)) &

                          ((hdu[1].data['WAVELENGTH'] > lyman[1])|
                           (hdu[1].data['WAVELENGTH'] < lyman[0])) &
                          ((hdu[1].data['WAVELENGTH'] > oxygen[1]) |
                           (hdu[1].data['WAVELENGTH'] < oxygen[0]))
                          )[0]

    return data_index

#-------------------------------------------------------------------------------

def get_both_filenames(filename):
    """ Get a list of both filenames for FUV data

    Regardless if rootname_corrtag_a.fits or rootname_corrtag_b.fits
    is passed in, both will be returned in a list.

    Parameters
    ----------
    filename : str
        full path to COS file

    Returns
    -------
    files : tuple
        rootname_corrtag_a.fits, rotname_corrtag_b.fits

    """

    if '_a.fits' in filename:
        other_filename = filename.replace('_a.fits', '_b.fits')
    elif '_b.fits' in filename:
        other_filename = filename.replace('_b.fits', '_a.fits')
    else:
        raise ValueError("filename doesn't match FUV convention".format(filename))

    filename_list = [filename, other_filename]
    filename_list.sort()

    return (filename_list[0], filename_list[1])

#-------------------------------------------------------------------------------

def get_extraction_region(hdu, segment, mode='spectrum'):
    """Get y_start,y_end for given extraction

    """

    if mode == 'spectrum':
        height =  hdu[1].header['SP_HGT_%s' % (segment)]
        location = hdu[1].header['SP_LOC_%s' % (segment)]

    elif mode == 'background1':
        height =  hdu[1].header['B_HGT1_%s' % (segment)]
        location = hdu[1].header['B_BKG1_%s' % (segment)]

    elif mode == 'background2':
        height =  hdu[1].header['B_HGT2_%s' % (segment)]
        location = hdu[1].header['B_BKG2_%s' % (segment)]

    y_start = location - height//2
    y_end = location + height//2

    return y_start, y_end

#-------------------------------------------------------------------------------

def get_fluxes(hdu, index):
    """ Return response curve and wavelength range for specified mode
    """

    fluxtab = hdu[0].header.get('FLUXTAB', '')

    fluxfile = expand_refname(fluxtab)

    if not len(index):
        return np.ones(hdu['events'].data['time'].shape)[index]

    if (not fluxfile) or (not os.path.exists(fluxfile)):
        print(' WARNING: Fluxfile not available %s,' % fluxfile )
        print(' using unity flux calibration instead.')
        return np.ones(hdu['events'].data['time'].shape)[index]


    flux_hdu = fits.open(fluxfile)
    setting_index = np.where((flux_hdu[1].data['SEGMENT'] == hdu[0].header['segment']) &
                             (flux_hdu[1].data['OPT_ELEM'] == hdu[0].header['opt_elem']) &
                             (flux_hdu[1].data['CENWAVE'] == hdu[0].header['cenwave']) &
                             (flux_hdu[1].data['APERTURE'] == hdu[0].header['aperture']))[0]
    if len(setting_index) == 0:
        print('No row in fluxtab found for this dataset, no FLUXCAL performed')
        return np.ones(hdu['events'].data['time'].shape)[index]
    elif len(setting_index) > 1:
        raise ValueError('Too many rows found: {}'.format(len(setting_index)))


    resp_wave = flux_hdu[1].data[setting_index]['WAVELENGTH'].flatten()
    response = flux_hdu[1].data[setting_index]['SENSITIVITY'].flatten()

    data_max = hdu['events'].data['wavelength'][index].max()
    data_min = hdu['events'].data['wavelength'][index].min()

    if data_max > resp_wave.max():
        print("Expanding minumum response curve by {}".format(data_max - resp_wave.max()))
        resp_wave[np.argmax(resp_wave)] = data_max

    if data_min < resp_wave.min():
        print("Expanding minimum response curve by {}".format(data_min - resp_wave.min()))
        resp_wave[np.argmin(resp_wave)] = data_min

    interp_func = interp1d(resp_wave, response)
    all_resp = interp_func(hdu['events'].data['wavelength'][index])

    return all_resp

#-------------------------------------------------------------------------------

def get_tds(hdu, index):
    """ return tds correction for each event
    """

    try:
        tdstab = hdu[0].header['TDSTAB']
    except KeyError:
        tdstab = ''


    tdsfile = expand_refname(tdstab)

    if not len(index):
        return np.ones(hdu['events'].data['time'].shape)[index]

    if (not tdsfile) or (not os.path.exists(tdsfile)):
        print(' WARNING: tdsfile not available %s,' % tdsfile)
        print(' using unity TDS correction instead.')
        return np.ones(hdu[ 'events' ].data['time'].shape)[index]

    tds_data = fits.getdata(tdsfile, ext=1)
    REF_TIME = fits.getval(tdsfile, 'REF_TIME', ext=1)
    mode_index = np.where((tds_data['OPT_ELEM'] == hdu[0].header['opt_elem']) &
                          (tds_data['APERTURE'] == hdu[0].header['aperture']) &
                          (tds_data['SEGMENT'] == hdu[0].header['segment']))[0]


    if len(mode_index) == 0:
        print('No row in tdstab found for this dataset, no TDSCORR performed')
        return np.ones(hdu['events'].data['time'].shape)[index]
    elif len(mode_index) > 1:
        raise ValueError('Too many rows found: {}'.format(len(mode_index)))

    mode_line = tds_data[mode_index]

    tds_nt = mode_line['NT'][0]
    tds_wavelength = mode_line['WAVELENGTH'][0]
    smaller_index = np.where(mode_line['TIME'][0][:tds_nt] < hdu[1].header['EXPSTART'])[0]
    time_index = smaller_index.max()

    tds_slope = mode_line['SLOPE'][0]
    tds_intercept = mode_line['INTERCEPT'][0]

    calculate_drop = lambda time, ref_time, slope, intercept: \
        (((time - ref_time) * slope) / (365.25*100)) + intercept

    correction_array = np.zeros(len(tds_wavelength))
    for i, wave in enumerate(tds_wavelength):
        correction = calculate_drop(hdu[1].header['EXPSTART'],
                                    REF_TIME,
                                    tds_slope[time_index,i],
                                    tds_intercept[time_index,i])
        correction_array[i] = correction

    #-------
    # interpolate onto input arrays
    #------

    interp_function = interp1d(tds_wavelength, correction_array, 1)
    response_correction = interp_function(hdu[1].data['wavelength'][index])

    return response_correction

#-------------------------------------------------------------------------------



"""Library of I/O routines to get data into a LightCurve object.

"""

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import os
import scipy
import numpy as np
from datetime import datetime
import astropy
from astropy.io import fits
from astropy.table import Table, vstack
from astropy.time import Time

from .version import version as __version__
from .cos import extract as extract_cos
from .cos import get_both_filenames
from .stis import extract as extract_stis
from .utils import is_uniq

__all__ = ['check_filetype',
           'read',
           'composite',
           'prepare_header']

#-------------------------------------------------------------------------------

def check_filetype(filename):
    """Determine the type of data being input.

    File type is determined by the culumns in the first data extension.

    Parameters
    ----------
    filename : str
        name of the input file

    Returns
    -------
    filetype : str
        determined type of file

    """

    corrtag_names = set(['TIME',
                         'RAWX',
                         'RAWY',
                         'XCORR',
                         'YCORR',
                         'XDOPP',
                         'XFULL',
                         'YFULL',
                         'WAVELENGTH',
                         'EPSILON',
                         'DQ',
                         'PHA'])

    lightcurve_names = set(['TIMES',
                            'MJD',
                            'GROSS',
                            'COUNTS',
                            'NET',
                            'FLUX',
                            'FLUX_ERROR',
                            'BACKGROUND',
                            'ERROR'])

    tag_names = set(['TIME',
                     'AXIS1',
                     'AXIS2',
                     'DETAXIS1'])

    stis_corrtag_names = set(['TIME',
                              'RAWX',
                              'RAWY',
                              'XCORR',
                              'YCORR',
                              'WAVELENGTH',
                              'EPSILON',
                              'DQ'])

    with fits.open(filename) as hdu:
        input_names = set([item.upper() for
                           item in hdu[1].data.names])

    if input_names == corrtag_names:
        filetype = 'cos_corrtag'
    elif input_names == lightcurve_names:
        filetype = 'lightcurve'
    elif input_names == tag_names:
        filetype = 'stis_tag'
    elif input_names == stis_corrtag_names:
        filetype = 'stis_corrtag'
    else:
        filetype = None

    return filetype

#-------------------------------------------------------------------------------

def read(source=None, **kwargs):
    verbosity = kwargs.get('verbosity', 0)

    if isinstance(source, dict):
        data = source
        meta = {}

    elif source == None:
        if verbosity: print("Initializing empty table.")

        data = {'dataset': np.array([]),
                'times': np.array([]),
                'mjd': np.array([]),
                'bins': np.array([]),
                'gross': np.array([]),
                'background': np.array([]),
                'flux': np.array([]),
                'counts': np.array([]),
                'error': np.array([]),
                'flux_error': np.array([]),
                'net': np.array([]),
                'signal_to_noise': np.array([])}

        meta = {'filename': None}

    else:
        filetype = check_filetype(source)
        if verbosity: print("Found {} as input filetype:".format(filetype))

        if filetype == 'lightcurve':
            return Table.read(source)
        elif filetype == 'cos_corrtag':
            data, meta = extract_cos(source, **kwargs)
        elif filetype == 'stis_tag' or filetype == 'stis_corrtag':
            data, meta = extract_stis(source, **kwargs)
        else:
            raise IOError("Filetype not recognized: {}".format(filetype))

        meta['outname'] = source[:9] + '_curve.fits'
        meta['filetype'] = filetype

        meta['GEN_DATE'] = (str(datetime.now()), 'Creation Date')
        meta['LC_VER'] = (__version__, 'lightcurve version used')
        meta['AP_VER'] = (astropy.__version__, 'Astropy version used')
        meta['NP_VER'] = (np.__version__, 'Numpy version used')
        meta['SP_VER'] = (scipy.__version__, 'Scipy version used')

        meta['expstart'] = data['mjd'].min()
        meta['expend'] = data['mjd'].max()
        meta['exptime'] = data['bins'].sum()
        meta['stepsize'] = (meta['stepsize'], 'Bin size (seconds)')

        meta['WMIN'] = (meta['wlim'][0], 'Minimum wavelength extracted')
        meta['WMAX'] = (meta['wlim'][1], 'Maximum wavelength extracted')
        meta['XMIN'] = (meta['xlim'][0], 'Minimum x-coordinate extracted')
        meta['XMAX'] = (meta['xlim'][1], 'Maximum x-coordinate extracted')

        if meta['instrument'] == 'COS':
            if 'ylima' in meta:
                meta['YMIN_A'] = (meta['ylima'][0], 'Minimum y-coordinate extracted from FUVA')
                meta['YMAX_A'] = (meta['ylima'][1], 'Maximum y-coordinate extracted from FUVA')
            if 'ylimb' in meta:
                meta['YMIN_B'] = (meta['ylimb'][0], 'Minimum y-coordinate extracted from FUVB')
                meta['YMAX_B'] = (meta['ylimb'][1], 'Maximum y-coordinate extracted from FUVB')
        elif meta['instrument'] == 'STIS':
            meta['YMIN'] = (meta['ylim'][0], 'Minimum y-coordinate extracted')
            meta['YMAX'] = (meta['ylim'][1], 'Maximum y-coordinate extracted')
        else:
            raise ValueError("{} not a recognized instrument".format(meta['instrument']))

    data['counts'] = data['gross'] - data['background']
    data['error'] = np.sqrt(data['gross'] + data['background'])
    data['flux_error'] = np.sqrt(data['gross'] + data['background']) / (data['gross'] - data['background']) * data['flux']
    data['net'] = (data['gross'] - data['background']) / data['bins']
    data['signal_to_noise'] = data['gross'] / np.sqrt(data['gross'] + data['background'])

    return Table(data, meta=meta)

#-------------------------------------------------------------------------------

def quicklook(filename):
    """ Quick plotting function for extracted lightcurves
    """

    with fits.open(filename) as hdu:

        fig = plt.figure()
        ax = fig.add_subplot(1, 1, 1)

        ax.plot(hdu[1].data['times'], hdu[1].data['gross'], 'o')

        ax.set_xlabel('Time (s)')
        ax.set_ylabel('Gross Counts')

        fig.suptitle(filename)
        fig.savefig(filename.replace('.fits', '.png'))
        plt.close(fig)

#-------------------------------------------------------------------------------

def composite(filelist, output, trim=True, **kwargs):
    """Creates a composite lightcurve from files in filelist and saves
    it to the save_loc.

    Parameters
    ----------
    filelist : list
        A list of full paths to the input files.
    output : string
        The path to the location in which the composite lightcurve is saved.
    trim : bool, opt
        Trim wavelengths to common ranges for all files
    """

    print("Creating composite lightcurve from:")
    print("\n".join(filelist))
    override = {}

    for i, filename in enumerate(filelist):
        with fits.open(filename) as hdu:
            dq = hdu[1].data['DQ']
            wave = hdu[1].data['wavelength']
            xcorr = hdu[1].data['xcorr']
            ycorr = hdu[1].data['ycorr']
            sdqflags = hdu[1].header['SDQFLAGS']

            if (hdu[0].header['INSTRUME'] == "COS") and (hdu[0].header['DETECTOR'] == 'FUV'):
                #-- if using COS FUV, these are the good limits
                HARD_WMIN = 912
                HARD_WMAX = 1800

                other_file = [item for item in get_both_filenames(filename) if not item == filename][0]
                if os.path.exists(other_file):
                    with fits.open(other_file) as hdu_2:
                        dq = np.hstack([dq, hdu_2[1].data['DQ']])
                        wave = np.hstack([wave, hdu_2[1].data['wavelength']])
                        xcorr = np.hstack([xcorr, hdu_2[1].data['xcorr']])
                        ycorr = np.hstack([ycorr, hdu_2[1].data['ycorr']])
                        sdqflags |= hdu_2[1].header['SDQFLAGS']
            else:
                #-- Super wide for NUV
                HARD_WMIN = 912
                HARD_WMAX = 9000

            index = np.where((np.logical_not(dq & sdqflags)) &
                             (wave > HARD_WMIN) &
                             (wave < HARD_WMAX) &
                             (xcorr >= 0) &
                             (ycorr >= 0))

            if not len(index[0]):
                print('No Valid events')
                continue

            if i == 0:
                wmax = wave[index].max()
                wmin = wave[index].min()
            else:
                wmax = min(wmax, wave[index].max())
                wmin = max(wmin, wave[index].min())
                print(wmin, '-->', wmax)


    if trim:
        print('Using wavelength range of: {}-{}'.format(wmin, wmax))
        kwargs['wlim'] = (wmin, wmax)

    override['wmin'] = wmin
    override['wmax'] = wmax

    all_lc = []
    for filename in filelist:
        tmp_lc = read(filename, **kwargs)
        if np.any(tmp_lc['gross'] == 0):
            continue
        else:
            all_lc.append(tmp_lc)

    for i, lc in enumerate(all_lc):
        if i == 0:
            out_lc = lc
        else:
            lc['dataset'] = i+1
            out_lc = vstack([out_lc, lc], metadata_conflicts='warn')

    if output.endswith('.fits') or output.endswith('.fits.gz'):
        for key in list(out_lc.meta.keys()):
            if len(key) > 8:
                print("Deleting key {} from output header:".format(key))
                del out_lc.meta[key]

    out_lc.write(output)

    prepare_header(output, filelist, override)

#-------------------------------------------------------------------------------

def prepare_header(filename, filelist, override={}):
    """Prepare headers with MAST requirements"""
    telescop = set()
    instrume = set()
    detector = set()
    filter = set()

    for i, exposure in enumerate(filelist):
        with fits.open(exposure) as hdu:
            print(exposure)
            telescop.add(hdu[0].header['TELESCOP'])
            instrume.add(hdu[0].header['INSTRUME'])
            detector.add(hdu[0].header['DETECTOR'])
            filter.add(hdu[0].header['OPT_ELEM'])

            if i == 0:
                ra_targ = hdu[0].header['ra_targ']
                dec_targ = hdu[0].header['dec_targ']
                equinox = hdu[0].header['equinox']
                targname = hdu[0].header['targname']
                tardescr = hdu[0].header.get('TARDESCR', '')
                tardesc2 = hdu[0].header.get('TARDESC2', '')

    with fits.open(filename, mode='update') as hdu:
        #-- HSLP keywords
        hdu[0].header['PROPOSID'] = 13902
        hdu[0].header['HLSPLEAD'] = 'Justin C. Ely'
        hdu[0].header['PR_INV_L'] = 'Ely'
        hdu[0].header['PR_INV_F'] = 'Justin'
        hdu[0].header['PR_INV_M'] = 'Charles'
        hdu[0].header['HLSPNAME'] = 'The Lightcurve Legacy of COS and STIS'
        hdu[0].header['HLSPACRN'] = 'LLOCS'
        hdu[0].header['CITATION'] = ''

        hdu[0].header['RA_TARG'] = ra_targ
        hdu[0].header['DEC_TARG'] = dec_targ
        hdu[0].header['EQUINOX'] = equinox
        hdu[0].header['TARGNAME'] = targname
        hdu[0].header['TARDESCR'] = tardescr
        hdu[0].header['TARDESC2'] = tardesc2
        hdu[0].header['EXTNAME'] = 'PRIMARY'
        hdu[1].header['EXTNAME'] = 'LIGHTCURVE'

        uniq, value = is_uniq(telescop)
        hdu[0].header['telescop'] = value

        uniq, value = is_uniq(instrume)
        hdu[0].header['instrume'] = value
        if not uniq:
            for i, val in enumerate(list(instrume)):
                hdu[0].header['instru{:0>2}'.format(1)] = val

        uniq, value = is_uniq(detector)
        hdu[0].header['detector'] = value
        if not uniq:
            for i, val in enumerate(list(detector)):
                hdu[0].header['detect{:0>2}'.format(1)] = val


        uniq, value = is_uniq(filter)
        hdu[0].header['filter'] = value
        if not uniq:
            for i, val in enumerate(list(filter)):
                hdu[0].header['filter{:0>2}'.format(1)] = val

        hdu[1].header['WMIN'] = override.get('wmin')
        hdu[1].header['WMAX'] = override.get('wmax')

        hdu[0].header['DATE-OBS'] = Time(hdu[1].data['MJD'].min(), format='mjd').iso
        hdu[0].header['EXPSTART'] = hdu[1].data['mjd'].min()
        hdu[0].header['EXPEND'] = hdu[1].data['mjd'].max()
        hdu[0].header['EXPTIME'] = hdu[1].data['bins'].sum()

#-------------------------------------------------------------------------------



"""
Tests for extraction of COS data.

"""
import numpy as np
from astropy.io import fits
import os

from .. import io
from ..cos import extract_index, get_both_filenames

#-------------------------------------------------------------------------------

def generate_test_files(outname='test_corrtag_a.fits', epsilon=1):
    hdu_out = fits.HDUList(fits.PrimaryHDU())
    hdu_out[0].header['detector'] = 'FUV'
    hdu_out[0].header['segment'] = 'FUVA'
    hdu_out[0].header['OBSTYPE'] = 'SPECTROSCOPIC'

    image = np.zeros((1024, 16384))
    image[100] = 1

    y_coords, x_coords = np.where(image > 0)
    n_events = len(y_coords)
    exptime = 100.0

    time_col = fits.Column('time', 'D', 'time', array=np.linspace(0, exptime, n_events))
    rawx_col = fits.Column('rawx', 'I', 'MJD', array=x_coords)
    rawy_col = fits.Column('rawy', 'I', 'MJD', array=y_coords)
    xcorr_col = fits.Column('xcorr', 'I', 'MJD', array=x_coords)
    ycorr_col = fits.Column('ycorr', 'I', 'MJD', array=y_coords)
    xdopp_col = fits.Column('xdopp', 'I', 'counts', array=x_coords)
    xfull_col = fits.Column('xfull', 'I', 'counts', array=x_coords)
    yfull_col = fits.Column('yfull', 'I', 'counts', array=y_coords)
    wavelength_col = fits.Column('wavelength', 'I', 'counts/s', array=np.ones(n_events) * 1200)
    epsilon_col = fits.Column('epsilon', 'D', 'ergs/s', array=np.ones(n_events) * epsilon)
    dq_col = fits.Column('dq', 'I', 'cnts', array=np.zeros(n_events))
    pha_col = fits.Column('pha', 'I', 'counts', array=np.ones(n_events) * 14)

    tab = fits.BinTableHDU.from_columns([time_col,
                                         rawx_col,
                                         rawy_col,
                                         xcorr_col,
                                         ycorr_col,
                                         xdopp_col,
                                         xfull_col,
                                         yfull_col,
                                         wavelength_col,
                                         epsilon_col,
                                         dq_col,
                                         pha_col])
    hdu_out.append(tab)

    hdu_out[1].header['EXTNAME'] = 'EVENTS'
    hdu_out[1].header['exptime'] = exptime
    hdu_out[1].header['expstart'] = 56000
    for segment in ['a', 'b']:
        hdu_out[1].header['sp_hgt_{}'.format(segment)] = 20
        hdu_out[1].header['sp_loc_{}'.format(segment)] = 100
        hdu_out[1].header['b_hgt1_{}'.format(segment)] = 20
        hdu_out[1].header['b_bkg1_{}'.format(segment)] = 150
        hdu_out[1].header['b_hgt2_{}'.format(segment)] = 20
        hdu_out[1].header['b_bkg2_{}'.format(segment)] = 50

    hdu_out[1].header['sdqflags'] = 16

    hdu_out.writeto(outname, clobber=True)

#-------------------------------------------------------------------------------

class test_extraction:
    def setUp(self):
        self.test_file = 'test_corrtag_a.fits'
        generate_test_files(outname=self.test_file)

    def tearDown(self):
        os.remove(self.test_file)

    def test_int_step(self):
        for stepsize in [.1, .5, 1, 2, 5, 10]:
            lc = io.read(self.test_file, step=stepsize)
            assert lc['gross'].sum() == 16384, "Extraction didn't find all the counts, step={}".format(stepsize)

    def test_step_truncation(self):
        for stepsize in [.3, 3]:
            lc = io.read(self.test_file, step=stepsize)
            assert (lc['gross'].sum() < 16384), "Extraction didn't truncate properly, step={}".format(stepsize)

#-------------------------------------------------------------------------------

class test_epsilon:
    def setUp(self):
        self.test_file = 'test_corrtag_a.fits'
        generate_test_files(outname=self.test_file, epsilon=1.25)

    def tearDown(self):
        os.remove(self.test_file)

    def test_simple_epsilon(self):
        lc = io.read(self.test_file)
        assert lc['gross'].sum() == 16384 * 1.25, 'Espilon not accounted for {}'.format(lc['gross'].sum())

#-------------------------------------------------------------------------------

def test_extract_index():
    """test that the extracted indexes works"""

    test_file = 'dq_corrtag_a.fits'
    generate_test_files(outname=test_file)

    with fits.open(test_file, mode='update') as hdu:
        hdu[1].data['dq'][:] = 8

        good_index = extract_index(hdu, 0, 16384, 0, 1024, -1, 10000, 8)
        assert len( good_index ) == 0, 'Should be no good data: {}'.format( len(good_index) )

        good_index = extract_index( hdu, 0, 16384, 0, 1024, -1, 10000, 8376)
        assert len( good_index ) == 0, 'Still should be no good data: {}'.format( len(good_index) )

    os.remove(test_file)

#-------------------------------------------------------------------------------

def test_get_both_filenames():
    """make sure i can find both"""

    test_names = ('rootname_corrtag_a.fits', 'rootname_corrtag_b.fits')

    assert get_both_filenames(test_names[0]) == test_names, \
        "Failed trying to find the B data with no path"

    assert get_both_filenames(test_names[1]) == test_names, \
        "Failed trying to find the A data with no path"

    test_names = ('/home/ely/rootname_corrtag_a.fits',
                  '/home/ely/rootname_corrtag_b.fits')

    assert get_both_filenames(test_names[0]) == test_names, \
        "Failed trying to find the B data with a path"

    assert get_both_filenames(test_names[1]) == test_names, \
        "Failed trying to find the A data with a path"



"""
Tests for the LightCurve Object itsself.

"""
from astropy.table import Table
import numpy as np

from .. import io

#-------------------------------------------------------------------------------

class test_derived_columns:
    def setUp(self):
        self.data = {'times': np.arange(10),
                     'mjd': np.arange(10),
                     'bins': np.ones(10),
                     'gross': np.ones(10)*5,
                     'background': np.ones(10)*1,
                     'flux': np.ones(10)*1}

        self.lc = io.read(self.data)

    def tearDown(self):
        pass

    def test_counts(self):
        assert np.array_equal(self.lc['counts'], np.ones(10)*4), \
            'Counts not calculated right'

    def test_net(self):
        assert np.array_equal(self.lc['net'], np.ones(10)*4), \
            'Net not calculated right'

    def test_error(self):
        assert np.array_equal(self.lc['error'], np.ones(10) * np.sqrt(6)), \
            'Error not calculated right with no background'

#-------------------------------------------------------------------------------



from ..stis import dqinit, map_image
#from ..stis_calib import map_image

import numpy as np

def test_map_image():
    """Test the cython extension"""


    dq = np.random.random_integers(0, 10, 100)
    x, y = np.meshgrid(np.arange(10), np.arange(10))
    x = x.flatten()
    y = y.flatten()
    dq_im = dq.reshape((10, 10))

    assert np.array_equal(dq, map_image(dq_im, x, y)), "Mapping failed with ints"

    dq = np.random.random_sample(100) * 10
    dq_im = dq.reshape((10, 10))
    assert np.array_equal(dq, map_image(dq_im, x, y)), "Mapping failed with floats"



from scipy.signal import lombscargle

__all__ = ['lomb']

#-------------------------------------------------------------------------------

def lomb(time, counts, frequencies):
    """Compute the lombscargle periodigram

    Necessary wrapper around the set lomscargle algorithm
    https://github.com/scipy/scipy/issues/2643

    Parameters
    ----------
    time : np.ndarray
        array of data times
    counts : np.ndarray
        array of counts
    frequencies : np.ndarray
        What frequencies

    Returns
    -------
    freqs : np.ndarray
        calculated freqencies

    """

    time = time.byteswap().newbyteorder().astype('float64')
    counts = counts.byteswap().newbyteorder().astype('float64')

    freqs = lombscargle(time, counts, frequencies)

    return freqs

#-------------------------------------------------------------------------------



from setuptools import setup, find_packages

setup(
    name = 'lightcurve',
    url = 'http://justincely.github.io/lightcurve/',
    version = '0.6.2',
    description = 'Create lightcurves from HST/COS and HST/STIS data',
    author = 'Justin Ely',
    author_email = 'ely@stsci.edu',
    keywords = ['astronomy'],
    classifiers = ['Programming Language :: Python',
                   'Programming Language :: Python :: 3',
                   'Development Status :: 1 - Planning',
                   'Intended Audience :: Science/Research',
                   'Topic :: Scientific/Engineering :: Astronomy',
                   'Topic :: Scientific/Engineering :: Physics',
                   'Topic :: Software Development :: Libraries :: Python Modules'],
    packages = find_packages(),
    install_requires = ['astropy',
                        'numpy>=1.9',
                        'scipy',
                        'numba>=0.24.0',
                        'llvmlite>=0.9.0',
                        'nose',
                        'six'],
    scripts =  ['scripts/lightcurve'],
    )



"""General purpose utility functions"""

import os

__all__ = ['expand_refname',
           'enlarge',
           'is_uniq']

#-------------------------------------------------------------------------------

def expand_refname(refname):
    '''Expand header reference file name to full path if $ is present.

    Parameters
    ----------
    refname, str
        reference file name

    Returns
    -------
    reffile, str
        expanded full path to reference file

    '''

    if '$' in refname:
        refpath, reffile = refname.split('$')

        try:
            reffile = os.path.join(os.environ[refpath], reffile)
        except KeyError:
            pass

    else:
        refpath = './'
        reffile = refname

    return reffile

#-------------------------------------------------------------------------------

def enlarge(a, x=2, y=None):
    """Enlarges 2D image array a using simple pixel repetition in both dimensions.
    Enlarges by factor x horizontally and factor y vertically.
    If y is left as None, uses factor x for both dimensions."""

    assert a.ndim == 2

    if y == None:
        y = x

    for factor in (x, y):
        assert factor.__class__ == int
        assert factor > 0

    return a.repeat(y, axis=0).repeat(x, axis=1)

#-------------------------------------------------------------------------------

def is_uniq(values):
    """ Check if input items are unique

    Parameters
    ----------
    values : set
        set of all values

    Returns
    -------
    True/False, MULTI/unique value
    """

    if len(values) == 0:
        return True, ''
    elif len(values) == 1:
        return True, list(values)[0]
    else:
        return False, 'MULTI'

#-------------------------------------------------------------------------------



""" Utility functions for extracting STIS spectral data into lightcurves

"""

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import os
import numpy as np
import scipy
from scipy.interpolate import interp1d
from datetime import datetime
from numba import jit
import astropy
from astropy.io import fits as fits

from .utils import expand_refname, enlarge
from .cos import extract_index, calc_npixels
from .version import version as  __version__

__all__ = ['extract',
           'stis_corrtag',
           'map_image',
           'epsilon',
           'dqinit']

#-------------------------------------------------------------------------------

def extract(filename, **kwargs):
    """ Extract lightcurve from STIS dataset

    This is the main driver of the lightcuve extracion, and definitely
    needs some better documentation.

    Parameters
    ----------
    filename : str
        name of FITS file to extract from
    **kwargs : dict
        arbitrary keyword arguements for tailored extraction

    Kwarg parameters
    ----------------
    verbosity : int, default=0
        Verbosity level for print output
    step : int, default=1
        timestep in seconds for output Lightcurve
    wlim : tuple

    Returns
    -------
    data, meta : Astropy table, dict
        Table with extracted data and dictionary of metadata pairs
    """

    verbosity = kwargs.get('verbosity', 0)
    step = kwargs.get('step', 1)
    wlim = kwargs.get('wlim', (2, 10000))
    xlim = kwargs.get('xlim', (0, 2048))
    ylim = kwargs.get('ylim', (0, 2048))
    filter_airglow = kwargs.get('filter_airglow', True)

    hdu = fits.open(filename)
    if fits.getval(filename, 'OBSTYPE') == 'IMAGING':
        print("Imaging observation found, resetting limits.")
        xlim = (0, 2048)
        ylim = (0, 2048)
        wlim = (-1, 1)

    input_headers = {'a':{}}

    for i, ext in enumerate(hdu):
        try:
            input_headers['a'][i] = ext.header._cards
        except AttributeError:
            pass

    meta = {'source': filename,
            'instrument' : 'STIS',
            'headers': input_headers,
            'stepsize': step,
            'wlim': wlim,
            'xlim': xlim,
            'ylim': ylim}

    if verbosity:
        print('#-----------------------------------------#')
        print('Running LightCurve extraction for STIS Data')
        print('#-----------------------------------------#')
        print()
        print('Extracting from: {}'.format(filename))
        print('With arguments:')
        print('step : {}'.format(step))
        print('wlim : {}'.format(wlim))
        print('xlim : {}'.format(xlim))
        print('ylim : {}'.format(ylim))
        print('filter_airglow : {}'.format(filter_airglow))
        print()

    SECOND_PER_MJD = 1.15741e-5

    time = hdu[1].data['time']
    #time = np.array([round(val, 3) for val in hdu[1].data['time']]).astype(np.float64)

    if not len(time):
        end = 0
        start = 0
    else:
        end = min(time.max(), hdu[1].header['EXPTIME'])
        start = time.min()

    all_steps = np.arange(start, end+step, step)

    if all_steps[-1] > end:
        truncate = True
    else:
        truncate = False

    ### need to fix sdqflags
    index = extract_index(hdu,
                          xlim[0], xlim[1],
                          ylim[0], ylim[1],
                          wlim[0], wlim[1],
                          hdu[1].header['SDQFLAGS'],
                          filter_airglow=filter_airglow)

    if verbosity:
        print("#{} events".format(len(index)))

    gross = np.histogram(hdu['events'].data['time'][index],
                         all_steps,
                         weights=hdu['events'].data['epsilon'][index])[0]

    n_pixels = calc_npixels(hdu, index, xlim)

    response_array = get_fluxes(hdu, index).mean()
    weights = hdu['events'].data['epsilon'][index] / step / response_array
    print("WARNING: The flux is a lie")
    flux =  np.histogram(hdu['events'].data['time'][index], all_steps, weights=weights)[0] / n_pixels

    print("WARNING: The background is a lie")
    background = np.zeros(gross.shape)
    mjd = hdu[1].header['EXPSTART'] + np.array(all_steps[:-1]) * SECOND_PER_MJD
    bins = np.ones(len(gross)) * step
    times = all_steps[:-1]

    if truncate:
        if verbosity:
            print('Truncating the last event bin')

        gross = gross[:-1]
        flux = flux[:-1]
        background = background[:-1]
        mjd = mjd[:-1]
        bins = bins[:-1]
        times = times[:-1]

    data = {'dataset': np.ones(times.shape),
            'times': times,
            'mjd': mjd,
            'bins': bins,
            'gross': gross,
            'background': background,
            'flux': flux}

    if verbosity:
        print('Finished extraction for {}'.format(filename))
        print()

    return data, meta

#-------------------------------------------------------------------------------

def stis_corrtag(tagfile, clean=True):
    """Create a COS-like corrtag file for STIS data

    Parameters
    ----------
    tagfile, str
        input STIS time-tag data file

    """
    print("Creating STIS corrtag for {}".format(tagfile))
    if '_tag.fits' in tagfile:
        x1d_filename = tagfile.replace('_tag.fits', '_x1d.fits')
    elif '_corrtag.fits' in tagfile:
        x1d_filename = tagfile.replace('_corrtag.fits', '_x1d.fits')

    with fits.open(tagfile, 'readonly') as hdu:
        n_events = len(hdu[1].data['TIME'])

        time_data = hdu[1].data['TIME']

        if 'DETAXIS1' in hdu[1].data.names:
            #-- Y data is same in raw and corrected coordinates
            rawx_data = hdu[1].data['DETAXIS1']
            rawy_data = hdu[1].data['AXIS2']
            xcorr_data = hdu[1].data['AXIS1']
            ycorr_data = hdu[1].data['AXIS2']
        else:
            rawx_data = hdu[1].data['RAWX']
            rawy_data = hdu[1].data['RAWY']
            xcorr_data = hdu[1].data['XCORR'].astype(np.int32)
            ycorr_data = hdu[1].data['YCORR'].astype(np.int32)

        #-- copy over the primary
        header0 = hdu[0].header.copy()
        header1 = hdu[1].header.copy()

    eps_data = epsilon(tagfile)
    dq_data = dqinit(tagfile)
    if header0['OBSTYPE'] == 'IMAGING':
        wave_data = np.zeros(n_events)

    elif not os.path.exists(x1d_filename):
        print("Could not find associated extracted spectrum {}".format(x1d_filename))
        wave_data = np.ones(n_events) * hdu[0].header['CENTRWV']

    elif header0['OPT_ELEM'].startswith('E'):
        #-- Put in average wavelength for now
        with fits.open(x1d_filename) as x1d:
            #--initialize everything to 0 wavelength
            wave_data = np.zeros(n_events)

            offset = header1['shifta2']
            for order in x1d[1].data:
                spec_center = order['a2center'] + offset
                spec_height = order['extrsize']

                index = np.where((ycorr_data < spec_center+spec_height) &
                                 (ycorr_data > spec_center-spec_height))[0]

                int_pix = integerize_pixels(xcorr_data[index])
                wave_data[index] = order['wavelength'][int_pix]

    else:
        #-- Grab wavelengths from the x1d file
        int_pix = integerize_pixels(xcorr_data)
        with fits.open(x1d_filename) as x1d:
            wave_data = x1d[1].data['wavelength'][0][int_pix]

    #-- Writeout corrtag file
    hdu_out = fits.HDUList(fits.PrimaryHDU())

    hdu_out[0].header['GEN_DATE'] = (str(datetime.now()), 'Creation Date')
    hdu_out[0].header['LC_VER'] = (__version__, 'lightcurve version used')
    hdu_out[0].header['AP_VER'] = (astropy.__version__, 'Astropy version used')
    hdu_out[0].header['NP_VER'] = (np.__version__, 'Numpy version used')
    hdu_out[0].header['SP_VER'] = (scipy.__version__, 'Scipy version used')

    hdu_out[0].header.extend(header0, end=True)

    time_col = fits.Column('time', 'D', 'second', array=time_data)
    rawx_col = fits.Column('rawx', 'D', 'pixel', array=rawx_data)
    rawy_col = fits.Column('rawy', 'D', 'pixel', array=rawy_data)
    xcorr_col = fits.Column('xcorr', 'D', 'pixel', array=xcorr_data)
    ycorr_col = fits.Column('ycorr', 'D', 'pixel', array=ycorr_data)
    epsilon_col = fits.Column('epsilon', 'D', 'factor', array=eps_data)
    wave_col = fits.Column('wavelength', 'D', 'angstrom', array=wave_data)
    dq_col = fits.Column('dq', 'I', 'DQ', array=dq_data)

    tab = fits.new_table([time_col,
                            rawx_col,
                            rawy_col,
                            xcorr_col,
                            ycorr_col,
                            epsilon_col,
                            wave_col,
                            dq_col])
    hdu_out.append(tab)

    hdu_out[1].header.extend(header1, end=True)

    hdu_out.writeto(tagfile.replace('_tag.fits', '_corrtag.fits'), clobber=True)
    if clean and '_tag.fits' in tagfile:
        print("Removing input tagfile")
        os.remove(tagfile)

#-------------------------------------------------------------------------------

def integerize_pixels(xcoords):

    int_pix = np.round(xcoords).astype(np.int32)
    int_pix = np.where(int_pix < 0, 0, int_pix)
    int_pix = np.where(int_pix > 2047, 2047, int_pix)
    int_pix //= 2

    return int_pix

#-------------------------------------------------------------------------------

@jit
def map_image(image, xcoords, ycoords, default=0):
    n_coord = len(xcoords)
    out_vals = np.zeros(n_coord)

    for i in range(n_coord):
        x = xcoords[i]
        y = ycoords[i]

        if x < 0 or x >= 2048:
            val = default
        elif y < 0 or y >= 2048:
            val = default
        else:
            val = image[y, x]

        out_vals[i] = val

    return out_vals

#-------------------------------------------------------------------------------

def epsilon(tagfile):
    """Compute the total epsilon factor for each event

    Compute the flatfield correction from the P-flat and L-flat reference files
    (PFLTFILE and LFLTFILE respectively).

    Parameters
    ----------
    tagfile, str
        input STIS time-tag data file

    Returns
    -------
    epsilon, np.ndarray
        array of epsilons

    """

    print("Calculating Epsilon")

    with fits.open(tagfile) as hdu:

        epsilon_out = np.ones(hdu[1].data['time'].shape)

        #-- Flatfield correction
        for ref_flat in ['PFLTFILE', 'LFLTFILE']:
            reffile = expand_refname(hdu[0].header[ref_flat])
            print('FLATFIELD CORRECTION {}: {}'.format(ref_flat, reffile))

            if not os.path.exists(reffile):
                print("{} not found, correction not performed".format(reffile))
                return np.ones(len(hdu[1].data))

            with fits.open(reffile) as image_hdu:
                image = image_hdu[1].data

                if not image.shape == (2048, 2048):
                    x_factor = 2048 // image.shape[1]
                    y_factor = 2048 // image.shape[0]

                    print('Enlarging by {},{}'.format(x_factor, y_factor))
                    image = enlarge(image, x_factor, y_factor)

                #--indexing is 1 off
                if 'AXIS1' in hdu[1].data.names:
                    epsilon_out *= map_image(image,
                                             hdu[1].data['AXIS1'] - 1,
                                             hdu[1].data['AXIS2'] - 1)
                else:
                    epsilon_out *= map_image(image,
                                             hdu[1].data['XCORR'].astype(np.int32) - 1,
                                             hdu[1].data['YCORR'].astype(np.int32) - 1)

    return epsilon_out

#-------------------------------------------------------------------------------

def dqinit(tagfile):
    """Compute the data quality information for each pixel from the BPIXTAB.

    Parameters
    ----------
    tagfile, str
        input STIS time-tag data file

    Returns
    -------
    dq, np.ndarray
        array of bitwise dq flags

    """

    print("Calculating DQ")

    with fits.open(tagfile) as hdu:

        reffile = expand_refname(hdu[0].header['BPIXTAB'])
        print('BPIXTAB used {}'.format(reffile))

        if not os.path.exists(reffile):
            print("{} not found, correction not performed".format(reffile))
            return np.zeros(len(hdu[1].data))

        with fits.open(reffile) as bpix:
            #-- Mama bpix regions are in lo-res pixels
            dq_im = np.zeros((1024, 1024)).astype(np.int32)

            #-- indexing is 1 off
            for line in bpix[1].data:
                lx = line['PIX1'] - 1
                ly = line['PIX2'] - 1
                dpix = line['LENGTH']
                axis = line['AXIS']
                flag = line['VALUE']

                if axis == 1:
                    #-- Along X
                    dq_im[ly, lx:lx + dpix] |= flag
                elif axis == 2:
                    #-- Along Y
                    dq_im[ly:ly+dpix, lx] |= flag

            #-- Needs to be expanded into Hi-res
            dq_im = enlarge(dq_im, 2, 2)


        #-- Map to the events
        #-- indexing is 1 off
        if 'AXIS1' in hdu[1].data.names:
            dq_out = map_image(dq_im,
                               hdu[1].data['AXIS1'] - 1,
                               hdu[1].data['AXIS2'] - 1)
        else:
            dq_out = map_image(dq_im,
                               hdu[1].data['XCORR'].astype(np.int32) - 1,
                               hdu[1].data['YCORR'].astype(np.int32) - 1)

    return dq_out

#-------------------------------------------------------------------------------

def crazy_histogram2d(x, y, bins=(2048, 2048)):
    """  Bin events to a 2d image.

    This is faster than the numpy version.

    http://stackoverflow.com/questions/8805601/efficiently-create-2d-histograms-from-large-datasets

    Parameters
    ----------
    x : array-like
        x-coordinates to bin
    y : array-like
        y-coordinates to bin
    bins : tuple, optional
        bins of output image

    Returns
    -------
    grid, x_bins, y_bins : (np.ndarray, np.ndarray, np.ndarray)
        binned image, x bins, and y bins
    """

    try:
        nx, ny = bins
    except TypeError:
        nx = ny = bins

    xmin, xmax = x.min(), x.max()
    ymin, ymax = y.min(), y.max()
    dx = (xmax - xmin) / (nx - 1.0)
    dy = (ymax - ymin) / (ny - 1.0)

    weights = np.ones(x.size)

    # Basically, this is just doing what np.digitize does with one less copy
    xyi = np.vstack((x,y)).T
    xyi = np.floor(xyi, xyi).T

    # Now, we'll exploit a sparse coo_matrix to build the 2D histogram...
    grid = scipy.sparse.coo_matrix((weights, xyi), shape=(nx, ny)).toarray().T

    return grid, np.linspace(xmin, xmax, nx), np.linspace(ymin, ymax, ny)

#-------------------------------------------------------------------------------

def get_fluxes(hdu, index):
    """ Return response curve and wavelength range for specified mode
    """

    fluxtab = hdu[0].header.get('FLUXTAB', '')

    fluxfile = expand_refname(fluxtab)

    if not len(index):
        return np.ones(hdu['events'].data['time'].shape)[index]

    if (not fluxfile) or (not os.path.exists(fluxfile)):
        print(' WARNING: Fluxfile not available %s,' % fluxfile )
        print(' using unity flux calibration instead.')
        return np.ones(hdu['events'].data['time'].shape)[index]


    flux_hdu = fits.open(fluxfile)
    setting_index = np.where((flux_hdu[1].data['SPORDER'] == hdu[0].header['segment']) &
                             (flux_hdu[1].data['OPT_ELEM'] == hdu[0].header['opt_elem']) &
                             (flux_hdu[1].data['CENWAVE'] == hdu[0].header['cenwave']))[0]

    if len(setting_index) == 0:
        print('No row in fluxtab found for this dataset, no FLUXCAL performed')
        return np.ones(hdu['events'].data['time'].shape)[index]
    elif len(setting_index) > 1:
        raise ValueError('Too many rows found: {}'.format(len(setting_index)))


    resp_wave = flux_hdu[1].data[setting_index]['WAVELENGTH'].flatten()
    response = flux_hdu[1].data[setting_index]['SENSITIVITY'].flatten()

    data_max = hdu['events'].data['wavelength'][index].max()
    data_min = hdu['events'].data['wavelength'][index].min()

    if data_max > resp_wave.max():
        print("Expanding minumum response curve by {}".format(data_max - resp_wave.max()))
        resp_wave[np.argmax(resp_wave)] = data_max

    if data_min < resp_wave.min():
        print("Expanding minimum response curve by {}".format(data_min - resp_wave.min()))
        resp_wave[np.argmin(resp_wave)] = data_min

    interp_func = interp1d(resp_wave, response)
    all_resp = interp_func(hdu['events'].data['wavelength'][index])

    return all_resp

#-------------------------------------------------------------------------------



""" Test utility functions in util module

"""

import os
import numpy as np

from ..utils import expand_refname, is_uniq, enlarge

from ..cos import expand_refname, extract_index

#-------------------------------------------------------------------------------

def test_expand_refname():
    """  Test that i can accurately find the header reference file """

    if 'lref' in os.environ: del os.environ['lref']
    assert expand_refname('lref$something_tds.fits') == 'something_tds.fits', \
        "Without lref, didn't return None"

    os.environ['lref'] = '/grp/hst/cdbs/lref/'
    full_refname = os.path.join( os.environ['lref'], 'something_tds.fits')
    assert expand_refname( 'lref$something_tds.fits' ) == full_refname, \
        "With lref, didn't find the file"

    assert expand_refname( '' ) == '', "didn't return blank on blank"

#-------------------------------------------------------------------------------

def test_is_uniq():
    empty_set = set()
    uniq_set = set([1])
    non_uniq_set = set([1, 2, 3])

    assert is_uniq(empty_set) == (True, ''), "Empty set failed"

    assert is_uniq(uniq_set) == (True, 1), "Unique set failed"

    assert is_uniq(non_uniq_set) == (False, 'MULTI'), "heterogeneous set failed"

#-------------------------------------------------------------------------------

def test_enlarge():
    small_array = np.ones((4, 4))
    big_array = np.ones((8, 8))

    assert np.array_equal(enlarge(small_array), big_array), "array enlarging failed"

#-------------------------------------------------------------------------------
